/*
 *  linux/fs/u2fs/inode.c
 *
 *  Copyright (C) 1994 Yossi Gottlieb (yogo@math.tau.ac.il)
 *
 *  Based on code and information:
 *	BSD code,  Copyright (c) 1982, 1986 Regents of the University of California.
 *		   All rights reserved.
 *	linux/fs/minix	Copyright (C) 1991, 1992, 1993  Linus Torvalds
 *
 *  Copyright (C) 1995  Bruno Haible <haible@ma2s2.mathematik.uni-karlsruhe.de>
 *    1995-01-09	modularized
 *    1995-01-20	added endianness support
 *    1995-01-21	recognize NeXTstep diskettes
 *
 *  Copyright (C) 1996  Eric Delaunay <delaunay@lix.polytechnique.fr>
 *    1996-03-06        cosmetic changes: split file into super.c & inode.c
 */

/* #include <linux/module.h> */

#include <linux/fs.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/locks.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#include <linux/u2fs_fs.h>
#include <linux/u2fs_fs_sb.h>
#include "u2fs.h"


void u2fs_read_inode(struct inode *inode)
{
	struct buffer_head *bh;
	struct u2fs_sb_info *sb = &inode->i_sb->u.u2fs_sb;
	struct dinode *dinode;
	kdev_t dev = inode->i_dev;
	int ino = inode->i_ino;
	int block;

	if (!(bh = bread(dev, sb->fs_block_base + fsbtodb(sb, ino_to_fsba(sb, ino)) + (ino_to_fsbo(sb, ino)/NIPB), BLOCK_SIZE))) {
		printk("u2fs: Unable to read inode from dev %s\n", kdevname(dev));
		return;
	}

	dinode = ((struct dinode *) bh->b_data) + ino_to_fsbo(sb, ino) % NIPB;
	inode->i_mode = dtohs(sb,dinode->di_mode);
	if (sb->fs_freebsd2) {
		inode->i_uid = dtohl(sb,dinode->di_u2.di_new.di_uid);
		inode->i_gid = dtohl(sb,dinode->di_u2.di_new.di_gid);
	} else {
		inode->i_uid = dtohs(sb,dinode->di_u1.di_oldids.di_uid);
		inode->i_gid = dtohs(sb,dinode->di_u1.di_oldids.di_gid);
	}
	inode->i_nlink = dtohs(sb,dinode->di_nlink);
	inode->i_size = dtohq(sb,dinode->di_qsize);
	inode->i_atime = dtohl(sb,dinode->di_atime);
	inode->i_mtime = dtohl(sb,dinode->di_mtime);
	inode->i_ctime = dtohl(sb,dinode->di_ctime);
	inode->i_blocks = dtohl(sb,dinode->di_blocks);

	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		inode->i_rdev = 0; /* XXX figure out where dev number is stored */
	else {
		if (DFASTLINK(inode))
			memcpy(inode->u.u2fs_i.i_un.i_usymlink, dinode->di_un.di_usymlink, MAXFASTLINK+1);
		else {
			for (block = 0; block<NDADDR; block++)
				inode->u.u2fs_i.i_un.i_addr.i_udb[block] =
					dtohl(sb,dinode->di_un.di_addr.di_udb[block]);
			for (block = 0; block<NIADDR; block++)
				inode->u.u2fs_i.i_un.i_addr.i_uib[block] =
					dtohl(sb,dinode->di_un.di_addr.di_uib[block]);
		}
	}
	brelse(bh);

	/* register file/inode operations table according to inode type */
	if (S_ISREG(inode->i_mode))
		inode->i_op = &u2fs_file_inode_operations;
	else
	if (S_ISDIR(inode->i_mode))
		inode->i_op = &u2fs_dir_inode_operations;
	else
	if (S_ISLNK(inode->i_mode))
		inode->i_op = &u2fs_symlink_inode_operations;
	else
	if (S_ISFIFO(inode->i_mode))
		init_fifo(inode);
}

/*
 * Reading whole u2fs blocks is impossible, they're too big. Fragments however,
 * are good enough (512/1024 bytes).
 *
 *
 *    u2fs_blknum()
 *
 *    translates a logical fragment number (for file) into absolute
 *    fragment number (based on BSD's ufs_bmap).
 */
daddr_t u2fs_blknum(struct inode *inode, int bn)
{
	struct buffer_head *bh;
	int nb;	/* disk block */
	struct u2fs_sb_info *sb = &inode->i_sb->u.u2fs_sb;
	int i, j, sh;
	kdev_t dev = inode->i_sb->s_dev;
	int fragofs = fragnum(sb, bn); /* where in block fragment is located */
	u2fs_daddr_t *bap;

	bn = fragstoblks(sb, bn);
	if (bn < NDADDR) {
		nb = inode->u.u2fs_i.i_un.i_addr.i_udb[bn];
		if (nb == 0)
			return -1;
		return nb + fragofs;
	}

	/*
	 * Determine the number of levels of indirection.
	 */
	sh = 1;
	bn -= NDADDR;
	for (j = NIADDR; j > 0; j--) {
		sh *= NINDIR(sb);
		if (bn < sh)
			break;
		bn -= sh;
	}
	if (j == 0)
		return -1;
	/*
	 * Fetch through the indirect blocks.
	 */
	nb = inode->u.u2fs_i.i_un.i_addr.i_uib[NIADDR - j];
	if (nb == 0)
		return -1;
	for (; j <= NIADDR; j++) {
		sh /= NINDIR(sb);
		i = (bn / sh) % NINDIR(sb);

		if (!(bh = bread(dev, sb->fs_block_base + fsbtodb(sb, nb+(i/DNINDIR(sb))), BLOCK_SIZE)))
			return -1;

		bap = (u2fs_daddr_t *)bh->b_data;
		nb = dtohl(sb, bap[i % DNINDIR(sb)]);
		brelse(bh);

		if (nb == 0)
			return -1;
	}

	return nb + fragofs;
}

/*
 * u2fs_bread(), u2fs_getblk() -- does what the name implies, on specified
 * inode's fragment.
 */

struct buffer_head *u2fs_bread(struct inode *inode, int bn)
{
	register int blk = u2fs_blknum(inode, bn);
	return blk == -1 ? NULL : bread(inode->i_sb->s_dev, inode->i_sb->u.u2fs_sb.fs_block_base + fsbtodb(&inode->i_sb->u.u2fs_sb, blk) , BLOCK_SIZE);
}

struct buffer_head *u2fs_getblk(struct inode *inode, int bn)
{
	register int blk = u2fs_blknum(inode, bn);
	return blk == -1 ? NULL : getblk(inode->i_sb->s_dev, inode->i_sb->u.u2fs_sb.fs_block_base + fsbtodb(&inode->i_sb->u.u2fs_sb, blk) , BLOCK_SIZE);
}
